/*#############################################################################
 cortmag.c: make cortical magnification data files -- Marty Sereno
###############################################################################
93/10/13 -- first hack
14/10/08 -- format, ANSI-ify
17/09/25 -- option to print help
##############################################################################*/
/* compile: gcc -D{Linux,Darwin} cortmag.c -lm -o cortmag */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
/*#include <jv.h>*/      /* NeXT fixed transcendental mathlib */
#include <string.h>

/* color table */
#define  CORTMAG_BORDER   0.3   /* max ~0.25 mm/deg */
#define  PNTIMG_BORDER    1.0   /* max ~0.7 mm */
#define  CONVERT   (2.0*M_PI/360.0)      /* degrees times this ==> radians */

/* true for exact match between S and str */
#define  MATCH_STR(S)   (strncmp(str,S,strlen(S)) == 0)

/* function declarations */
int main(int argc, char **argv);
void print_help(void);


/*** begin main */
int main(int argc, char **argv)
{
  int    i, x, y, xi, yi, xsize, ysize, offset;
  float  radius_deg, theta_deg, dia_deg;
  float  point_img_size, cort_mag_factor;
  float  x_estimate, y_estimate;
  float  mm_per_pixel, dcort, dr;
  float  x1, x2, y1, y2;
  float  **radius_array, **theta_array, **dia_array;
  char   str[81];
  FILE   *fpcortmag, *fppntimg, *fpradius, *fptheta, *fpdia;

  if (argc!=6) {
    printf("\n");
    printf(
      "use: cortmag <rootfilename> <xsize> <ysize> <offset> <mm_per_pix>\n");
    printf("     cortmag -h,-help\n");
    printf("\n");
    if (argc==2) {
      strcpy(str,argv[1]);
      if ((MATCH_STR("-h")) || (MATCH_STR("-help"))) { print_help(); }
    }
    exit(0);
  }
  xsize = atoi(argv[2]);
  ysize = atoi(argv[3]);
  offset = atoi(argv[4]);
  mm_per_pixel = atof(argv[5]);

  /* open infiles */
  sprintf(str,"%s.radius.%dx%d",argv[1],xsize,ysize);
  fpradius = fopen(str,"r");
  if (fpradius==NULL) {
    printf("cortmag: ### couldn't find radius data file: %s\n",str); exit(1);}
  printf("cortmag: using datafile: %s\n",str);
  sprintf(str,"%s.theta.%dx%d",argv[1],xsize,ysize);
  fptheta = fopen(str,"r");
  if (fptheta==NULL) {
    printf("cortmag: ### couldn't find theta data file: %s\n",str); exit(1);}
  printf("cortmag: using datafile: %s\n",str);
  sprintf(str,"%s.dia.%dx%d",argv[1],xsize,ysize);
  fpdia = fopen(str,"r");
  if (fpdia==NULL) {
    printf("cortmag: ### couldn't find dia data file: %s\n",str); exit(1);}
  printf("cortmag: using datafile: %s\n",str);

  /* malloc */
  radius_array = (float **)malloc(ysize*sizeof(float *)); 
  for (i=0; i<ysize; i++)
    radius_array[i] = (float *)malloc(xsize*sizeof(float));
  theta_array = (float **)malloc(ysize*sizeof(float *));  
  for (i=0; i<ysize; i++)
    theta_array[i] = (float *)malloc(xsize*sizeof(float));
  dia_array = (float **)malloc(ysize*sizeof(float *));  
  for (i=0; i<ysize; i++)
    dia_array[i] = (float *)malloc(xsize*sizeof(float));

  /* read data */
  for (yi=0; yi<ysize; yi++)
  for (xi=0; xi<xsize; xi++) {
    fscanf(fpradius,"%d %d %f",&x,&y,&radius_deg);  /* OK if shuffled */
    radius_array[y][x] = radius_deg*CONVERT;
    fscanf(fptheta,"%d %d %f",&x,&y,&theta_deg);
    theta_array[y][x] = theta_deg*CONVERT;
    fscanf(fpdia,"%d %d %f",&x,&y,&dia_deg);
    dia_array[y][x] = dia_deg*CONVERT;
  }
  fclose(fpradius);
  fclose(fptheta);
  fclose(fpdia);

  /* open outfiles */
  sprintf(str,"%s.cortmag.%dx%d",argv[1],xsize,ysize);
  printf("cortmag: cortmag file: %s\n",str);
  fpcortmag = fopen(str,"w");
  sprintf(str,"%s.pntimg.%dx%d",argv[1],xsize,ysize);
  printf("cortmag: pntimage file: %s\n",str);
  fppntimg = fopen(str,"w");

  /* use three infiles to estimate cortmag, point image; write rasterfiles */
  for (y=0; y<ysize; y++)
  for (x=0; x<xsize; x++) {
    /* skip border points */
    if (x<offset || y<offset || x>=xsize-offset || y>=ysize-offset) {
      fprintf(fpcortmag,"%d\t%d\t%2.4f\n",x,y,CORTMAG_BORDER);
      fprintf(fppntimg,"%d\t%d\t%2.4f\n",x,y,PNTIMG_BORDER);
    }
    else {
      /* est. mag fact (mm/deg) in x,y dir (flat approx to spherepolar OK) */
      dcort = (2.0 * (float)offset) * mm_per_pixel;
      x1 = radius_array[y][x-offset] * cos(theta_array[y][x-offset]);
      y1 = radius_array[y][x-offset] * sin(theta_array[y][x-offset]);
      x2 = radius_array[y][x+offset] * cos(theta_array[y][x+offset]);
      y2 = radius_array[y][x+offset] * sin(theta_array[y][x+offset]);
      dr = hypot(x2-x1,y2-y1);
      x_estimate = dcort/(dr/CONVERT);  /* go back to degrees */
      x1 = radius_array[y-offset][x] * cos(theta_array[y-offset][x]);
      y1 = radius_array[y-offset][x] * sin(theta_array[y-offset][x]);
      x2 = radius_array[y+offset][x] * cos(theta_array[y+offset][x]);
      y2 = radius_array[y+offset][x] * sin(theta_array[y+offset][x]);
      dr = hypot(x2-x1,y2-y1);
      y_estimate = dcort/(dr/CONVERT);
      cort_mag_factor = (x_estimate + y_estimate)/2.0;
      fprintf(fpcortmag,"%d\t%d\t%2.4f\n",x,y,cort_mag_factor);

      /* minimum estimate of point image size (ignores RF jitter) */
      point_img_size = (dia_array[y][x]/CONVERT) * cort_mag_factor;
      fprintf(fppntimg,"%d\t%d\t%2.4f\n",x,y,point_img_size);
    }  /* all but edge */
  }  /* all grid points */
  fclose(fpcortmag);
  fclose(fppntimg);
  exit(0);
}

/*** begin functions */
void print_help()
{
printf("\
############################################################################\n\
              cortmag -- make cortical magnification data files\n\
\n\
                       written by Marty Sereno\n\
                  University of California, San Diego\n\
#############################################################################\n\
  Reads three ASCII rasterfiles of r,theta,dia data in degrees:\n\
\n\
    rootfilename.radius.30x25     format: x  y  radius(deg)\n\
    rootfilename.theta.30x25      format: x  y  theta(deg)\n\
    rootfilename.dia.30x25        format: x  y  dia(deg)\n\
\n\
    Calculates two new ASCII rasterfiles of local cortical magnification\n\
  (in mm cortex per degree of visual angle) and local point image\n\
  size (in mm: RFsize * cortical magnification):\n\
\n\
    rootfilename.cortmag.30x25    format: x  y  cortmag(mm/deg)\n\
    rootfilename.pntimg.30x25     format: x  y  pntimg(mm)\n\
\n\
    Command line specifies xsize, ysize, offset (for dx/dr).\n\
  An offset=1 means derivatives calculated across 3 pixels, offset=2\n\
  across 5 pixels, etc.  Averages estimates of cortical magnification\n\
  factor in x and y directions.  Writes <offset> wide BORDER-colored\n\
  pixels at edge of map (no gradient calculated there).\n\
#############################################################################\n\
\n\
");
}

