/*############################################################################
 Generate cpt file for GMT w/tksurfer color scale (in degrees)
##############################################################################
14/09/30 -- first hack
14/10/10 -- TODO: add 2 extra lines to clamp theta beyond -90/90
15/01/24 -- gamutflag for blue
############################################################################*/
/* compile: gcc gmtcpt.c -lm -o gmtcpt */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#define MATCH(A,B)        (!strcmp(A,B))
#define DEG2RAD   (M_PI/180.0)          /* deg*DEG2RAD => rad */
#define RAD2DEG   (180.0/M_PI)          /* rad*RAD2DEG => deg */

#define PAD_DEG    45    /* pad ends of polar scale */
#define BLUE_FRAC 0.88   /* -gamut: frac reduce blue for blue desaturate */
#define RG_FRAC   0.12   /* -gamut: frac increase R/G for blue desaturate */

/* type's */
#define ECCEN  0
#define POLAR  1 

int gamutflag=0;

int main(int argc, char **argv);
void set_complex_color(float x, float y, float angle_cycles, float angle_offset,
                       int revphaseflag, float *r, float *g, float *b);

int main(int argc, char **argv)
{
  int    i, type, cptlines;
  int    ri0, gi0, bi0;
  int    ri1, gi1, bi1;
  float  angcyc, angoff, revphase=0;
  float  x, y, a0, a1, adeg0, adeg1;
  float  r, g, b;

  if (argc<5) {
    printf("\n");
    printf("Use: gmtcpt <type:0=ecc,1=pol> <lines> <angcyc> <angoff> [-opt]\n");
    printf("\n");
    printf("  write tksurfer wheel GMT color pallete (*.cpt) to stdout\n");
    printf("  0=ecc: data range 0 to 90 deg\n");
    printf("  1=pol: data range -90 to 90 deg (clamp 45deg beyond ends\n");
    printf("  N.B.: both non-integer/end-clamped polar scale (fadef=0.99)\n");
    printf("\n");
    printf(" Options:\n");
    printf("\n");
    printf("  -gamut => desaturate blue (closer to print gamut)\n");
    printf("\n");
    printf(" Examples:\n");
    printf("\n");
    printf("   eccen: gmtcpt 0 25 1.9 -0.25\n");
    printf("   polar: gmtcpt 1 25 2.2 0.0\n");
    printf("\n");
    exit(0);
  }
  type = atoi(argv[1]);
  cptlines = atoi(argv[2]);
  angcyc = atof(argv[3]);
  angoff = atof(argv[4]);
  if (argc>5) {
    for (i=5; i<argc; i++) {
      if (MATCH(argv[i],"-gamut")) gamutflag = 1;
    }
  }

  if (angcyc==1.0) angcyc += 0.001;  /* always use clamping color scale */

  if (type==ECCEN) {
    revphase = 0;
    for (i=0; i<cptlines; i++) {
      a0 = (float)i * 2.0*M_PI/(float)cptlines;  /* colfunc inp: 0-2Pi */
      x = cos(a0);
      y = sin(a0);
      set_complex_color(x, y, angcyc, angoff, revphase, &r, &g, &b);
      ri0 = r*255.0;
      gi0 = g*255.0;
      bi0 = b*255.0;
      a1 = (float)(i+1) * 2.0*M_PI/(float)cptlines;
      x = cos(a1);
      y = sin(a1);
      set_complex_color(x, y, angcyc, angoff, revphase, &r, &g, &b);
      ri1 = r*255.0;
      gi1 = g*255.0;
      bi1 = b*255.0;
      adeg0 = (float)i     * 90.0/(float)cptlines;   /* cpt out: 0,90 deg */
      adeg1 = (float)(i+1) * 90.0/(float)cptlines;
      printf("%2.1f\t%d\t%d\t%d\t\t%2.1f\t%d\t%d\t%d\n",
              adeg0,ri0,gi0,bi0,adeg1,ri1,gi1,bi1);
    }
  }
  else if (type==POLAR) {
    revphase = 1;  /* default left hemifield */
    for (i=0; i<cptlines; i++) {
      a0 = 0.5*M_PI + (float)i * M_PI/(float)cptlines; /*colfunc inp: L hemif*/
      x = cos(a0);
      y = sin(a0);
      set_complex_color(x, y, angcyc, angoff, revphase, &r, &g, &b);
      ri0 = r*255.0;
      gi0 = g*255.0;
      bi0 = b*255.0;
      a1 = 0.5*M_PI + (float)(i+1) * M_PI/(float)cptlines;
      x = cos(a1);
      y = sin(a1);
      set_complex_color(x, y, angcyc, angoff, revphase, &r, &g, &b);
      ri1 = r*255.0;
      gi1 = g*255.0;
      bi1 = b*255.0;
      adeg0 = -90.0 + (float)i     * 180.0/(float)cptlines; /*out: -90,90deg*/
      adeg1 = -90.0 + (float)(i+1) * 180.0/(float)cptlines;
      if (i==0) {printf("%2.1f\t%d\t%d\t%d\t\t%2.1f\t%d\t%d\t%d\n",
                 adeg0-PAD_DEG,ri0,gi0,bi0,adeg0,ri0,gi0,bi0);}
      printf("%2.1f\t%d\t%d\t%d\t\t%2.1f\t%d\t%d\t%d\n",
              adeg0,ri0,gi0,bi0,adeg1,ri1,gi1,bi1);
      if (i==cptlines-1) {printf("%2.1f\t%d\t%d\t%d\t\t%2.1f\t%d\t%d\t%d\n",
                          adeg1,ri0,gi0,bi0,adeg1+PAD_DEG,ri1,gi1,bi1);}
    }
  }
  else { printf("gmtcpt: ### unrec type: %d (OK:0=eccen,1=polar)\n",type); }
  exit(0);
}

void set_complex_color(float x, float y, float angle_cycles, float angle_offset,
                       int revphaseflag, float *red, float *grn, float *blu)
{
  int    invphaseflag=0;
  float  fmid=0.1, fslope=3.0, fadef=0.99;
  float  f, fscale, a;
  float  r=0.5, g=0.5, b=0.5;
  float  tmpoffset=0.345, ipsiyellowfact=0.0;

  /* get amp (sqrtF, pval, prob-false-detect), optional rep|mask|mult|none) */
  f = sqrt(x*x + y*y);
  fscale = f;

  /* 1st soft-thresh+squash, amp vs. fthresh (else clamp amp 0-1) */
  if (fslope!=0.0) f = (1.0 + tanh(fslope*(f - fmid)))/2.0;
  else             f = (f<0.0)?0.0:((f>1.0)?1.0:f);

  /* get/modify/condition phase */
  a = atan2(y,x)/(2.0*M_PI);    /* -0.5 to 0.5 */
  if (revphaseflag) a = -a;
  if (invphaseflag) a += 0.5;
  a -= angle_offset;
  a = fmod(a,1.0);     /* -1 to 1 */
  if (a<0.0) a += 1.0; /*  0 to 1 */

  /* int or non-int color wheel */
  if (fmod(angle_cycles,1.0)==0.0) { /* integral cycles (eccen: RBG,wrap) */
    a -= 0.333333;           /* center on blue (1/3)*/
    a = angle_cycles*a;      /* allow multiple RBG cycles */
    a = fmod(a,1.0);
    if (a<0.0) a += 1.0;
    a = 3.0*a;
    r = g = b = 0.0;
    if (a<1.0)      { r = 1.0-a;       b = a; }             /* r->b */
    else if (a<2.0) { b = 1.0-(a-1.0); g = a-1.0; }         /* b->g */
    else            { r = a-2.0;       g = 1.0-(a-2.0); }   /* g->r */
  }
  else {  /* non-integral cycles (polar: RBG,no-wrap) */
    a -= 0.5;          /* center on blue (1/2) */
    a = angle_cycles*a;
    r = g = b = 0.0;
    if      (a<-0.33333) {r = 1.0;}                                /* r */
    else if (a<0.0)      {r = 1.0-(a-(-0.33333))/(0.0-(-0.33333)); /* r->b */
                          b = (a-(-0.33333))/(0.0-(-0.33333));}
    else if (a<0.33333)  {b = 1.0-(a)/(0.33333);                   /* b->g */
                          g = a/(0.33333);}
    else                 {g = 1.0;}                                /* g */
    if (gamutflag) {  /* hack blue closer to gamut */
      r += b*RG_FRAC*(1.0 - r);
      g += b*RG_FRAC*(1.0 - g);
      b *= BLUE_FRAC;
    }
    if (a > fadef*0.5*angle_cycles) {  /* g->gray (or yellow) */
      f = 1.0    -    (a-fadef*0.5*angle_cycles) /
               (0.5*angle_cycles - fadef*0.5*angle_cycles);
      r = (tmpoffset*(1.0 - f) + f*r);
      g = (tmpoffset*(1.0 - f) + f*g);
      b = (tmpoffset*(1.0 - f) + f*b);
      if (ipsiyellowfact != 0.0) { /*TODO yelfact: 2R/G vs 1B: less yellow*/
        r *= (1.0 - ipsiyellowfact) + ipsiyellowfact/tmpoffset;
        g += ipsiyellowfact*(1.0 - g);  /* add easier than adj slope */
        b *= (1.0 - ipsiyellowfact);
      }
    }
    if (a < -fadef*0.5*angle_cycles) {  /* r->gray (or yellow) */
      f = (a - (-0.5*angle_cycles)) /
        (0.5*angle_cycles - fadef*0.5*angle_cycles);
      r = (tmpoffset*(1.0 - f) + f*r);
      g = (tmpoffset*(1.0 - f) + f*g);
      b = (tmpoffset*(1.0 - f) + f*b);
      if (ipsiyellowfact != 0.0) {
        r += ipsiyellowfact*(1.0 - r);
        g *= (1.0 - ipsiyellowfact) + ipsiyellowfact/tmpoffset;
        b *= (1.0 - ipsiyellowfact);
      }
    }
  }
  *red = r = tmpoffset*(1.0 - fscale) + fscale*r;
  *grn = g = tmpoffset*(1.0 - fscale) + fscale*g;
  *blu = b = tmpoffset*(1.0 - fscale) + fscale*b;
}

