/*#############################################################################
 stretch.c: deform sparse x-y array of points -- Marty Sereno
###############################################################################
92/12/20 -- first hack
92/12/31 -- working: data,init,ave,orth energy terms
93/01/04 -- lin term; don't move w/energy=0; interpolated output
93/01/12 -- functions; write multiple pages using optional args
93/01/24 -- delete cross product hack
93/02/04 -- fix squared term scaling
93/02/09 -- fix linear truncation in higher power terms
93/02/23 -- smaller PostScript outfile w/defs
93/03/02 -- better structures
93/05/30 -- fixed NeXT mathlib (2.5x faster): cc stretch.c -ljv -lm -o stretch
93/08/17 -- annealing hack
04/02/17 -- test mesh^2, data+mesh^2, data+mesh^3 vs current mesh^3 (still best)
94/02/26 -- fix box
94/09/13 -- release mesh from data hack (for end)
96/09/23 -- min dist for move, fix min energy for move bug, float->double
14/10/08 -- format, ANSIify
17/09/25 -- option to print help
20/08/25 -- embed cmdline/parmfile as comments in ps outfile
  TODO: make mesh cell size parm (not xgridsize)
##########################################################################*/
/* compile: gcc -D{Linux,Darwin} stretch.c -lm -o stretch */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
/*#include <jv.h>*/    /* NeXT fixed 68040 mathlib transcendentals */
#include <string.h>

#ifdef IRIX
#  include <values.h>
#endif

#ifdef Darwin
#  include <i386/limits.h>    /* INT_MAX=2147483647 */
#  define MAXINT INT_MAX
#endif

/* default parms, adjust w/parm file */
#define  MARG            30.0   /* margin in points */
#define  THICKNESS       0.2    /* thickness of PostScript mesh lines */
#define  DATA_POINT      3.0    /* diameter of unfilled data point */
#define  LESION_POINT    7.0    /* diameter of filled lesion point */
#define  LANDSCAPE_FLAG    0    /* portrait mode */
#define  LEGEND_FLAG       1    /* with legend */
#define  SHOWMESH_FLAG     1    /* with mesh */
#define  SHOWSTART_FLAG    1    /* with initial points & line to final */
#define  SHOWFINAL_FLAG    1    /* with final points */
#define  DATA_MARG_FRAC   0.02  /* fraction of total grid beyond max x data*/
#define  SQFACTOR         1.0   /* frac squared term added dist,angles term */
#define  CROSSPNT         0.25  /* frac meshbox,90deg where sqr crosses lin*/
#define  X_GRIDSIZE       20    /* number of vertices in x direction */
#define  SET_DATA_DIST_TERM  1.0  /* coeff on data term in energy func */
#define  SET_INIT_DIST_TERM  0.01 /* coeff on init dist term in energy func */
#define  SET_AVE_DIST_TERM   0.1  /* coeff on ave dist term in energy func */
#define  SET_ORTH_TERM       0.1  /* coeff on orthog. term in energy func */
#define  SET_LIN_TERM        0.0  /* coeff on linear term in energy func */
#define  MOMENTUM            0.5  /* how much of last movement in current */
#define  EPSILON             0.1  /* small offset to test for min--in mm */
#define  FRAC_STEPS_TO_ANNEAL   0.0  /* fraction of total steps to anneal */
#define  FRAC_EPSILON_TO_ANNEAL 0.25 /* fraction of epsilon during annealing */
#define  FRAC_STEPS_TO_RELEASE  0.0  /* fraction total steps released */

/* other defs */
#define  ORIGBBOX          4.0   /* thickness of original bounding box */
#define  TOTAL_POINTS     2000   /* should re-malloc */
#define  LESION_POINTS     100   /* should re-malloc */
#define  MAX_NAMESIZE       11   /* OK */
#define  LEGEND_FONT       10.0
#define  DEBUG              0

#define  TOTAL_PYSIZE     792.0   /* (A4=842.0) page ysize in points */
#define  TOTAL_PXSIZE     612.0   /* (A4=595.0) page xsize in points */
#define  CONVERT  (M_PI/180.0) /* degrees times this ==> radians */
#define  EEPS  0.000000001         /* min energy for move */
#define  DEPS  0.000000001         /* min ddist (not needed for consistency)*/
#define  float double              /* need this  */

/* true for exact match between S and str */
#define  MATCH_STR(S)   (strncmp(str,S,strlen(S)) == 0)
#define  SQR(X)     ((X)*(X))
#define  CUBE(X)    ((X)*(X)*(X))
#define  FOURTH(X)  ((X)*(X)*(X)*(X))
#define  RAND_FLOAT(L,H)  ((L)+((float)random()/(float)MAXINT)*((H)-(L)))

typedef struct lesion_type_ {
  char   name[MAX_NAMESIZE];   /* name of lesions (from lesion file) */
  float  xstart;       /* starting location of lesion (from datafile) */
  float  ystart;
  int    xstart_vertex;/* number of mesh vertex nearest starting lesion loc*/
  int    ystart_vertex;
  float  xtarg;        /* target lesion location (from lesionfile) */
  float  ytarg;
  float  xtarg_mesh;   /* loc of mesh vertex that puts start lesion on targ */
  float  ytarg_mesh;
} lesion_type;

typedef struct rf_type_ {
  char   name[MAX_NAMESIZE];
  float  xcortex;
  float  ycortex;
  float  radius;
  float  theta;
  float  length;
  float  width;
  float  angle;
  float  xoffset;       /* offset from nearest vertex--to make outfile */
  float  yoffset;
  int    xvertex;       /* number of nearest vertex */
  int    yvertex;
  float  final_xcortex; /* interpolated stretched location */
  float  final_ycortex;
} rf_type;

typedef struct mesh_type_ {
  float  **xm;         /* mesh x coord for each x-y vertex */
  float  **ym;
  float  **xlastmove;  /* last x move for each x-y vertex */
  float  **ylastmove;
  int    x_gridsize;   /* parms */
  int    y_gridsize;
  float  x_incr;
  float  y_incr;
  float  sqfactor;
  float  crosspnt;
  float  data;         /* energy terms */
  float  init;
  float  ave;
  float  orth;
  float  lin;
  float  xmin;         /* PostScript */
  float  ymin;
  float  marg;
  float  x_center_marg;
  float  y_center_marg;
  float  mm_to_pts;
  float  thickness;
  float  data_point;
  float  lesion_point;
} mesh_type;

/* function declarations */
int main(int argc, char **argv);
void page_scale(float xsize, float ysize, int landscape_flag, float marg,
              float *x_center_marg, float *y_center_marg, float *units_to_pts);
void ps_header(FILE *fpout, int landscape_flag, int pages);
void ps_parms(FILE *fpout, int argc, char **argv);
void ps_nextpage(FILE *fpout, int nextpage);
void ps_trailer(FILE *fpout, int pages);
void ps_text(FILE *fpout, char *str, float x, float y, float fontsize);
void ps_pnt(FILE *fpout, float center_x, float center_y, float dia,
            float thickness, int filled_flag);
void ps_line(FILE *fpout, float x1, float y1, float x2, float y2,
             float thickness);
void ps_box(FILE *fpout, float x0, float y0, float x1, float y1,
            float thickness);
void ps_drawmesh(FILE *fpps, mesh_type mesh);
void ps_draworigmeshbox(FILE *fpps, float gx0, float gy0, float gx1, float gy1,
                        mesh_type mesh);
void ps_drawlesions(FILE *fpps, lesion_type *lesion, int num_lesion,
                    mesh_type mesh);
void ps_drawdata(FILE *fpps, rf_type *rf, int num_rf, mesh_type mesh,
                 int showstart_flag, int showfinal_flag);
void ps_drawlegend(FILE *fpps, char *filename, int landscape_flag, int steps,
                   float sqfactor, float crosspnt, float set_data_dist_term,
                   float set_init_dist_term, float set_ave_dist_term,
                   float set_orth_term, float set_lin_term, float momentum,
                   float epsilon, int maxsteps, int frac_steps_to_anneal,
                   float frac_epsilon_to_anneal, mesh_type mesh);
void interpolate_data(rf_type *rf, int num_rf, mesh_type mesh);
float **fmatrix(int ysize, int xsize);
void shuffle(int *array, int n);
void get_energy_terms(lesion_type *lesion, int num_lesion, mesh_type *mesh,
                      int x, int y, float xeps, float yeps, int release_flag);
double mydrem(double x, double y);
double mycopysign(double x, double y);
void print_help(void);


/*** begin main */
int main(int argc, char **argv)
{
  char   line[200], str[200], junk[5];

  int    i, j, k;
  int    x, y;
  int    matches;
  int    num_rf, num_lesion;
  int    x_gridsize, y_gridsize;
  int    landscape_flag, legend_flag;
  int    showmesh_flag, showstart_flag, showfinal_flag;
  int    *order;
  int    *frame;
  int    maxsteps, framecnt;
  int    release_flag;

  float  parm;
  float  thickness, data_point, lesion_point;
  float  sqfactor, crosspnt;
  float  ymax=0.0, xmax=0.0, ymin=0.0, xmin=0.0;
  float  xsize, ysize;
  float  total_xsize, total_ysize;
  float  marg;
  float  x_center_marg, y_center_marg;
  float  data_marg, data_marg_frac, y_cent_data_marg;
  float  mm_to_pts;
  float  xm, ym;
  float  xrem, yrem;
  float  momentum;
  float  epsilon;
  float  xcortex, ycortex;
  float  x_incr, y_incr;
  float  set_data_dist_term, set_init_dist_term, set_ave_dist_term;
  float  set_orth_term, set_lin_term;
  float  ex0, ex1, ey0, ey1;
  float  dex, dey;
  float  xeps, yeps;
  float  xmove, ymove;
  float  gx0, gx1, gy0, gy1;
  float  frac_steps_to_anneal, frac_epsilon_to_anneal, eps;
  float  frac_steps_to_release;
  float  **fmatrix();

  FILE   *fpparm, *fpdata, *fplesion, *fpstr, *fpps;

  lesion_type  lesion[LESION_POINTS];
  rf_type      rf[TOTAL_POINTS];
  mesh_type    mesh;

  if (argc<3) {
    printf("usage: %s <rootfilename> <steps> [steps] ...\n\n",argv[0]);
    printf("parms and   MARG = 30.0             X_GRIDSIZE = 20\n");
    printf("defaults:   THICKNESS = 0.2         SQFACTOR = 1.0\n");
    printf("            DATA_POINT = 3.0        CROSSPNT = 0.25\n");
    printf("            LESION_POINT = 7.0      SET_DATA_DIST_TERM = 1.0\n");
    printf("            LANDSCAPE_FLAG = 0      SET_INIT_DIST_TERM = 0.01\n");
    printf("            LEGEND_FLAG = 1         SET_AVE_DIST_TERM = 0.1\n");
    printf("            SHOWMESH_FLAG = 1       SET_ORTH_TERM = 0.1\n");
    printf("            SHOWSTART_FLAG = 1      SET_LIN_TERM = 0.0\n");
    printf("            SHOWFINAL_FLAG = 1      MOMENTUM = 0.5\n");
    printf("            DATA_MARG_FRAC = 0.02   EPSILON = 0.1\n");
    printf("            FRAC_STEPS_TO_ANNEAL = 0.0\n");
    printf("            FRAC_EPSILON_TO_ANNEAL = 0.25\n");
    printf("            FRAC_STEPS_TO_RELEASE = 0.0\n");
    printf("[version 19]\n");
    printf("\n");
    if (argc==2) {
      strcpy(str,argv[1]);
      if ((MATCH_STR("-h")) || (MATCH_STR("-help"))) { print_help(); }
    }
    exit(0);
  }

  /* try to open datafile (rootname.out) and lesionfile (rootname-les.out) */
  sprintf(str,"%s.out",argv[1]);
  fpdata = fopen(str,"r");
  if (fpdata==NULL) {
    printf("stretch: ### couldn't find main data file: %s\n",str);
    exit(1);
  }
  sprintf(str,"%s-les.out",argv[1]);
  fplesion = fopen(str,"r");
  if (fplesion==NULL)  {
    printf("stretch: ### couldn't find lesion data file: %s\n",str);
    exit(1);
  }

  /* extra printouts */
  maxsteps = 0;
  framecnt = argc - 2;
  frame = (int *)malloc(framecnt*sizeof(int));
  for (i=0; i<framecnt; i++)  {
    frame[i] = atoi(argv[i+2]);
    if (frame[i]<0) {printf("stretch: ### bad steps arg\n"); exit(1);}
    if (frame[i]>maxsteps) maxsteps = frame[i];
  }

  /* load default parms */
  marg = MARG;
  thickness = THICKNESS;
  data_point = DATA_POINT;
  lesion_point = LESION_POINT;
  landscape_flag = LANDSCAPE_FLAG;
  legend_flag = LEGEND_FLAG;
  showmesh_flag = SHOWMESH_FLAG;
  showstart_flag = SHOWSTART_FLAG;
  showfinal_flag = SHOWFINAL_FLAG;
  data_marg_frac = DATA_MARG_FRAC;
  x_gridsize = X_GRIDSIZE;
  sqfactor = SQFACTOR;
  crosspnt = CROSSPNT;
  set_data_dist_term = SET_DATA_DIST_TERM;
  set_init_dist_term = SET_INIT_DIST_TERM;
  set_ave_dist_term = SET_AVE_DIST_TERM;
  set_orth_term = SET_ORTH_TERM;
  set_lin_term = SET_LIN_TERM;
  momentum = MOMENTUM;
  epsilon = EPSILON;
  frac_steps_to_anneal = FRAC_STEPS_TO_ANNEAL;
  frac_epsilon_to_anneal = FRAC_EPSILON_TO_ANNEAL;
  frac_steps_to_release = FRAC_STEPS_TO_RELEASE;

  /* try to open parms file */
  fpparm = fopen("str.parm","r");
  if (fpparm == NULL)  {
    printf("stretch: using default parms; new parmfile is:  str.parm\n");
    fpparm = fopen("str.parm","w");
    fprintf(fpparm,"MARG = %3.2f\n",MARG);
    fprintf(fpparm,"THICKNESS = %3.2f\n",THICKNESS);
    fprintf(fpparm,"DATA_POINT = %3.2f\n",DATA_POINT);
    fprintf(fpparm,"LESION_POINT = %3.2f\n",LESION_POINT);
    fprintf(fpparm,"LANDSCAPE_FLAG = %d\n",LANDSCAPE_FLAG);
    fprintf(fpparm,"LEGEND_FLAG = %d\n",LEGEND_FLAG);
    fprintf(fpparm,"SHOWMESH_FLAG = %d\n",SHOWMESH_FLAG);
    fprintf(fpparm,"SHOWSTART_FLAG = %d\n",SHOWSTART_FLAG);
    fprintf(fpparm,"SHOWFINAL_FLAG = %d\n",SHOWFINAL_FLAG);
    fprintf(fpparm,"DATA_MARG_FRAC = %3.2f\n",DATA_MARG_FRAC);
    fprintf(fpparm,"X_GRIDSIZE = %d\n",X_GRIDSIZE);
    fprintf(fpparm,"SQFACTOR = %3.2f\n",SQFACTOR);
    fprintf(fpparm,"CROSSPNT = %3.2f\n",CROSSPNT);
    fprintf(fpparm,"SET_DATA_DIST_TERM = %3.4f\n",SET_DATA_DIST_TERM);
    fprintf(fpparm,"SET_INIT_DIST_TERM = %3.4f\n",SET_INIT_DIST_TERM);
    fprintf(fpparm,"SET_AVE_DIST_TERM = %3.4f\n",SET_AVE_DIST_TERM);
    fprintf(fpparm,"SET_ORTH_TERM = %3.4f\n",SET_ORTH_TERM);
    fprintf(fpparm,"SET_LIN_TERM = %3.4f\n",SET_LIN_TERM);
    fprintf(fpparm,"MOMENTUM = %3.2f\n",MOMENTUM);
    fprintf(fpparm,"EPSILON = %3.4f\n",EPSILON);
    fprintf(fpparm,"FRAC_STEPS_TO_ANNEAL = %3.4f\n",FRAC_STEPS_TO_ANNEAL);
    fprintf(fpparm,"FRAC_EPSILON_TO_ANNEAL = %3.4f\n",FRAC_EPSILON_TO_ANNEAL);
    fprintf(fpparm,"FRAC_STEPS_TO_RELEASE = %3.4f\n",FRAC_STEPS_TO_RELEASE);
    fclose(fpparm);
  }
  else {  /* read any parms in file; others leave default */
    printf("stretch: using **existing** parmfile:\n");
    while (fgets(line,80,fpparm) != NULL) { 
      if (sscanf(line,"%s %s %lf",str,junk,&parm) == 3)  {
        if      (MATCH_STR("MARG"))             marg = parm;
        else if (MATCH_STR("THICKNESS"))        thickness = parm;
        else if (MATCH_STR("DATA_POINT"))       data_point = parm;
        else if (MATCH_STR("LESION_POINT"))     lesion_point = parm;
        else if (MATCH_STR("LANDSCAPE_FLAG"))   landscape_flag = (int)parm;
        else if (MATCH_STR("LEGEND_FLAG"))      legend_flag = (int)parm;
        else if (MATCH_STR("SHOWMESH_FLAG"))    showmesh_flag = (int)parm;
        else if (MATCH_STR("SHOWSTART_FLAG"))   showstart_flag = (int)parm;
        else if (MATCH_STR("SHOWFINAL_FLAG"))   showfinal_flag = (int)parm;
        else if (MATCH_STR("DATA_MARG_FRAC"))   data_marg_frac = parm;
        else if (MATCH_STR("X_GRIDSIZE"))       x_gridsize = (int)parm;
        else if (MATCH_STR("SQFACTOR"))         sqfactor = parm;
        else if (MATCH_STR("CROSSPNT"))         crosspnt = parm;
        else if (MATCH_STR("SET_DATA_DIST_TERM")) set_data_dist_term = parm;
        else if (MATCH_STR("SET_INIT_DIST_TERM")) set_init_dist_term = parm;
        else if (MATCH_STR("SET_AVE_DIST_TERM")) set_ave_dist_term = parm;
        else if (MATCH_STR("SET_ORTH_TERM"))    set_orth_term = parm;
        else if (MATCH_STR("SET_LIN_TERM"))     set_lin_term = parm;
        else if (MATCH_STR("MOMENTUM"))         momentum = parm;
        else if (MATCH_STR("EPSILON"))          epsilon = parm;
        else if (MATCH_STR("FRAC_STEPS_TO_ANNEAL"))
                                                frac_steps_to_anneal = parm;
        else if (MATCH_STR("FRAC_EPSILON_TO_ANNEAL"))
                                               frac_epsilon_to_anneal = parm;
        else if (MATCH_STR("FRAC_STEPS_TO_RELEASE"))
                                                frac_steps_to_release = parm;
        else printf("stretch: don't know this parm: %s  ...ignored\n\n",str);
      }
      else {
        printf("stretch: ### couldn't parse this line in str.parm:  %s",line);
        printf("  ...quitting\n");
        exit(1);
      }
    }
    fclose(fpparm);
  } /* done reading parms */

  if (sqfactor > 1.0) {
    printf("stretch: ### reset sqfactor to 1.0\n");
    sqfactor = 1.0;
  }

  /*** get flattened cortex lesion file line; read 8 args data */
  i = num_lesion = 0;
  while (fgets(line,80,fplesion) != NULL) { 
    /* check if 8 arg line (w/ellipse coord)--toss RF data */
    if (!(sscanf(line,"%s %lf %lf %lf %lf %lf %lf %lf",lesion[i].name,
           &xcortex, &ycortex, &parm, &parm, &parm, &parm, &parm) == 8)) {
      printf("stretch: ### couldn't parse this line in lesion file:\n");
      printf("   %s",line);
      continue;
    }

    /* max,min over both data and lesion files */
    if (i==0)  { ymax=ymin=ycortex; xmax=xmin=xcortex; }
    if (xcortex>xmax) xmax = xcortex;
    if (ycortex>ymax) ymax = ycortex;
    if (xcortex<xmin) xmin = xcortex;
    if (ycortex<ymin) ymin = ycortex;

    /* store x and y */
    lesion[i].xtarg = xcortex;
    lesion[i].ytarg = ycortex;
    i++;
    if (i>=LESION_POINTS) {
      printf("stretch: ### overflowed tmp lesion mem (%d pts)\n",LESION_POINTS);
      exit(1);
    }
  }  /*** for all lines in lesion file */
  fclose(fplesion);
  num_lesion = i;
  if (DEBUG) {
    printf("stretch: lesion x(min,max) = (%3.2f,%3.2f) mm\n",xmin,xmax);
    printf("stretch: lesion y(min,max) = (%3.2f,%3.2f) mm\n",ymin,ymax);
  }
  /*** done with lesion file ***/

  /*** get datafile line; read 8 args data; look for x-y of lesion (by name) */
  i = num_rf = matches = 0;
  while (fgets(line,80,fpdata) != NULL) { 
    /* check if 8arg line (w/ellipse coord;if args bad, overwrites next time) */
    if (!(sscanf(line,"%s %lf %lf %lf %lf %lf %lf %lf",
                 str, &xcortex, &ycortex, &rf[i].radius, &rf[i].theta,
                 &rf[i].length, &rf[i].width, &rf[i].angle) == 8)) {
      printf("stretch: couldn't parse this line in data file:\n");
      printf("   %s",line);
      continue;
    }

    /* go through lesion names to see if lesion pnt; yes =>save x-y in lesion*/
    k = 0;
    for (j=0; j<num_lesion; j++) {
      if (MATCH_STR(lesion[j].name)) {
        k++;
        matches++;
        if (k>1) printf(
     "stretch: ### Warning: 2+ lesionfile lines match (%s) in datafile\n",str);
        lesion[j].xstart = xcortex;
        lesion[j].ystart = ycortex;
      }
    }
    strcpy(rf[i].name,str);

    /* maxs,mins (use initial maxs and mins from lesionfile) */
    if (xcortex>xmax) xmax = xcortex;
    if (ycortex>ymax) ymax = ycortex;
    if (xcortex<xmin) xmin = xcortex;
    if (ycortex<ymin) ymin = ycortex;

    /* store x and y */
    rf[i].xcortex = xcortex;
    rf[i].ycortex = ycortex;
    i++;
    if (i >= TOTAL_POINTS)  {
      printf("stretch: ### overflowed tmp data mem (%d points)\n",TOTAL_POINTS);
      exit(1);
    }
  }  /*** for all lines in file */
  fclose(fpdata);

  num_rf = i;
  if (matches!=num_lesion) {
    printf(
      "stretch: ### Warning: %d lesions but %d matches with datafile lines\n",
      num_lesion,matches);
    printf("  ...may have two datafile lesions with same name\n");
  }
  else printf("stretch: %d lesions matched to names in datafile\n",matches);
  /* overall size of data and lesion file */
  ysize = ymax - ymin;
  xsize = xmax - xmin;
  if (DEBUG) {
    printf("stretch: overall x(min,max) = (%3.2f,%3.2f) mm\n",xmin,xmax);
    printf("stretch: overall y(min,max) = (%3.2f,%3.2f) mm\n",ymin,ymax);
  }
  /*** done with data file ***/


  /*** get mesh parameters; save in struct; init mesh; save bounding box */
  /* in datafile units (mm)--remember: gridsize not zero-based */
  data_marg = data_marg_frac*xsize;
  total_xsize = xsize + 2.0*data_marg;
  total_ysize = ysize + 2.0*data_marg;

  /* make x go to edge (plus margin) */
  x_incr = y_incr = total_xsize/(float)(x_gridsize-1);

  /* center y (plus margin); ceil() gets pts inside; +1 for not zero-based */
  y_gridsize = ceil(total_ysize/y_incr) + 1;
  y_cent_data_marg = ((float)(y_gridsize-1)*y_incr - total_ysize)/2.0;

  /* save in mesh struct */
  mesh.x_gridsize = x_gridsize;
  mesh.y_gridsize = y_gridsize;
  mesh.x_incr = x_incr;
  mesh.y_incr = y_incr;
  mesh.sqfactor = sqfactor;
  mesh.crosspnt = crosspnt;

  /* save these too for convenience */
  mesh.marg = marg;
  mesh.thickness = thickness;
  mesh.data_point = data_point;
  mesh.lesion_point = lesion_point;

  /* malloc */
  printf("stretch: %d x %d mesh points\n",x_gridsize,y_gridsize);
  mesh.xm = fmatrix(y_gridsize,x_gridsize);
  mesh.ym = fmatrix(y_gridsize,x_gridsize);

  /* initialize a square mesh in mm */
  for (y=0; y<y_gridsize; y++)
  for (x=0; x<x_gridsize; x++) {
    mesh.xm[y][x] = -data_marg + xmin + x*x_incr;
    mesh.ym[y][x] = -data_marg - y_cent_data_marg + ymin + y*y_incr;
    if (DEBUG)  { /* introduce random jitter to test energy terms */
      mesh.xm[y][x] += RAND_FLOAT(-0.5,0.5);
      mesh.ym[y][x] += RAND_FLOAT(-0.5,0.5);
    }
  }

  /* save orig corners for bounding box */
  gx0 = mesh.xm[0][0];
  gy0 = mesh.ym[0][0];
  gx1 = mesh.xm[y_gridsize-1][x_gridsize-1];
  gy1 = mesh.ym[y_gridsize-1][x_gridsize-1];

  /*** find nearest-mesh-vertex numbers,offsets for data points at start */
  for (i=0; i<num_rf; i++) {
    xcortex = rf[i].xcortex - xmin + data_marg;  /* dist to mesh edge */
    xrem = drem(xcortex,x_incr); /* dist to vtx; drem: +/-rem nearest int*/
    if (xrem > 0.0)
      x = floor(xcortex/x_incr); /* get closest int */
    else
      x = ceil(xcortex/x_incr);  /* ditto--in right half of grid cell */
    rf[i].xoffset = xrem;        /* save offset for output */
    rf[i].xvertex = x;   /* save vtxnum near data point (or lesion start) */
    /* now y:  dist to mesh edge needs extra centering */
    ycortex = rf[i].ycortex - ymin + data_marg + y_cent_data_marg;
    yrem = drem(ycortex,y_incr);
    if (yrem > 0.0)
      y = floor(ycortex/y_incr);
    else
      y = ceil(ycortex/y_incr);
    rf[i].yoffset = yrem;
    rf[i].yvertex = y;

    /* chk if lesion; save nearest-vtx num again, use offset to get targ_mesh */
    for (j=0; j<num_lesion; j++)  {
      if (strncmp(lesion[j].name,rf[i].name,strlen(lesion[j].name)) == 0) {
        lesion[j].xstart_vertex = x;  /* copy vtx into lesion struct*/
        lesion[j].ystart_vertex = y;
        /* get offset targ location for mesh point nearest lesion start */
        /* NOTE: these assume square mesh */
        lesion[j].xtarg_mesh = lesion[j].xtarg - xrem; /* neg:undo offset */
        lesion[j].ytarg_mesh = lesion[j].ytarg - yrem;
      }
    }
  }  /* for all data points (includes untransformed lesions) */

  /*** ready to stretch ***/
  printf("stretch: doing %d total relaxation steps\n",maxsteps);
  printf("stretch: %d annealing steps at %2.1f epsilon\n",
        (int)((float)maxsteps*frac_steps_to_anneal),frac_epsilon_to_anneal);
  printf("stretch: %d steps released from data\n",
        (int)((float)maxsteps*frac_steps_to_release));
  printf("stretch:   sqfactor = %g\n",sqfactor);
  printf("stretch:   crosspnt = %g\n",crosspnt);
  printf("stretch: cell width = %2.3f mm\n",x_incr);
  printf("stretch:  data term = %2.1f\n",set_data_dist_term);
  printf("stretch:  init term = %g\n",set_init_dist_term);
  printf("stretch:   ave term = %g\n",set_ave_dist_term);
  printf("stretch:  orth term = %g\n",set_orth_term);
  printf("stretch:   lin term = %g\n",set_lin_term);
  printf("stretch:   momentum = %g\n",momentum);
  printf("stretch:    epsilon = %g mm\n",epsilon);
  mesh.xlastmove = fmatrix(y_gridsize,x_gridsize);
  mesh.ylastmove = fmatrix(y_gridsize,x_gridsize);
  for (y=0; y<y_gridsize; y++)
  for (x=0; x<x_gridsize; x++)
    mesh.xlastmove[y][x] = mesh.ylastmove[y][x] = 0.0;
  order = (int *)malloc(x_gridsize*y_gridsize*sizeof(int));
  for (i=0; i<x_gridsize*y_gridsize; i++)
    order[i] = i;

  /*** stretch loop: print PostScript(if arg), outfile(if last), then stretch*/
  srandom((int)1);
  for (i=0; i<maxsteps+1; i++)  {   /* +1 to get 0-step */

    /*** if PostScript needed this step: find max/min/xsize/ysize;scale;draw */
    for (j=0; j<framecnt; j++)  {
      if (i==frame[j]) {  /* check if we want to print--then center it */
        for (y=0; y<mesh.y_gridsize; y++)
        for (x=0; x<mesh.x_gridsize; x++)  {
          xm = mesh.xm[y][x];
          ym = mesh.ym[y][x];
          if (ym==0 && xm==0) { ymax=ymin=ym; xmax=xmin=xm; }
          if (xm>xmax) xmax = xm;
          if (ym>ymax) ymax = ym;
          if (xm<xmin) xmin = xm;
          if (ym<ymin) ymin = ym;
        }  /* for the whole mesh */
        ysize = ymax - ymin;
        xsize = xmax - xmin;
        page_scale(xsize,ysize,landscape_flag,mesh.marg,
                        &x_center_marg,&y_center_marg,&mm_to_pts);
        mesh.xmin = xmin;
        mesh.ymin = ymin;
        mesh.x_center_marg = x_center_marg;
        mesh.y_center_marg = y_center_marg;
        mesh.mm_to_pts = mm_to_pts;

        /* open file */
        sprintf(str,"%s.str.%d.ps",argv[1],i);
        fpps = fopen(str,"w");
        printf("\nstretch: PostScript outfile: %s\n",str);
        ps_header(fpps,landscape_flag,1);
        ps_parms(fpps,argc,argv);
        if (legend_flag)
          ps_drawlegend(fpps,argv[1],landscape_flag,i,sqfactor,crosspnt,
               set_data_dist_term,set_init_dist_term,set_ave_dist_term,
               set_orth_term,set_lin_term,momentum,epsilon,maxsteps,
               frac_steps_to_anneal,frac_epsilon_to_anneal,mesh);
        /* draw mesh; orig bounding box; lesions; data */
        if (showmesh_flag) ps_drawmesh(fpps,mesh);
        ps_draworigmeshbox(fpps,gx0,gy0,gx1,gy1,mesh); /* see how it moved */
        ps_drawlesions(fpps,lesion,num_lesion,mesh); /* start,targ:pts+mesh */
        interpolate_data(rf,num_rf,mesh);
        ps_drawdata(fpps,rf,num_rf,mesh,showstart_flag,showfinal_flag);
        ps_trailer(fpps,1);  /* just one page */
        fclose(fpps);
      }  /* if we're printing this frame */
    }  /* for checking all the desired frames */

    /*** write stretched output file for arrows.c, interpolate.c if last iter*/
    if (i==maxsteps) {
      sprintf(str,"%s.str",argv[1]);
      fpstr = fopen(str,"w");
      printf("stretch: stretched datafile: %s\n",str);
      interpolate_data(rf,num_rf,mesh);
      for (j=0; j<num_rf; j++)
        fprintf(fpstr,"%s\t%3.4f\t%3.4f\t%3.2f\t%3.2f\t%3.1f\t%3.1f\t%3.1f\n",
                rf[j].name,rf[j].final_xcortex,rf[j].final_ycortex, 
                rf[j].radius,rf[j].theta,rf[j].length,rf[j].width,rf[j].angle);
      fclose(fpstr);
    } /* if last */

    /*** one iteration on mesh (last wasted) */
    shuffle(order,x_gridsize*y_gridsize);
    if ((float)i < (float)maxsteps*(1.0 - frac_steps_to_anneal))
      eps = epsilon;
    else
      eps = epsilon*frac_epsilon_to_anneal;
    if ((float)i < (float)maxsteps*(1.0 - frac_steps_to_release))
      release_flag = 0;
    else
      release_flag = 1;
    for (j=0; j<x_gridsize*y_gridsize; j++) {
      k = order[j];
      y = k/x_gridsize;
      x = k%x_gridsize;

      /* estimate gradient in energy by checking in each direction */
      xeps = eps/2.0;  /* east */
      yeps = 0.0;
      get_energy_terms(lesion,num_lesion,&mesh,x,y,xeps,yeps,release_flag);
      ex0 = set_data_dist_term * mesh.data +
             set_init_dist_term * mesh.init + 
             set_ave_dist_term * mesh.ave +
             set_orth_term * mesh.orth +
             set_lin_term * mesh.lin;

      xeps = -eps/2.0;  /* west */
      yeps = 0.0;
      get_energy_terms(lesion,num_lesion,&mesh,x,y,xeps,yeps,release_flag);
      ex1 = set_data_dist_term * mesh.data +
             set_init_dist_term * mesh.init + 
             set_ave_dist_term * mesh.ave +
             set_orth_term * mesh.orth +
             set_lin_term * mesh.lin;

      xeps = 0.0;
      yeps = eps/2.0;  /* north */
      get_energy_terms(lesion,num_lesion,&mesh,x,y,xeps,yeps,release_flag);
      ey0 = set_data_dist_term * mesh.data +
             set_init_dist_term * mesh.init + 
             set_ave_dist_term * mesh.ave +
             set_orth_term * mesh.orth +
             set_lin_term * mesh.lin;

      xeps = 0.0;
      yeps = -eps/2.0;  /* south */
      get_energy_terms(lesion,num_lesion,&mesh,x,y,xeps,yeps,release_flag);
      ey1 = set_data_dist_term * mesh.data +
             set_init_dist_term * mesh.init + 
             set_ave_dist_term * mesh.ave +
             set_orth_term * mesh.orth +
             set_lin_term * mesh.lin;

      /*** move mesh to minimize energy and save step */
      dex = ex1 - ex0;  /* x energy diff */
      if (fabs(dex)<EEPS)
        xmove = 0.0;
      else
        xmove = copysign(eps/2.0,dex) + momentum*mesh.xlastmove[y][x];
      mesh.xm[y][x] += xmove;
      mesh.xlastmove[y][x] = xmove;

      dey = ey1 - ey0;
      if (fabs(dey)<EEPS)
        ymove = 0.0;
      else
        ymove = copysign(eps/2.0,dey) + momentum*mesh.ylastmove[y][x];
      mesh.ym[y][x] += ymove;
      mesh.ylastmove[y][x] = ymove;
    }  /* for all grid points */
    if (!(i%10) && i!=maxsteps)   {printf(".");      fflush(stdout);}
    if (!(i%100) && i!=maxsteps)  {printf(" %d ",i); fflush(stdout);}
  }  /*** for all iterations on mesh */
  exit(0);
}

/*** begin functions */
void page_scale(float xsize, float ysize, int landscape_flag, float marg,
                float *x_center_marg, float *y_center_marg, float *units_to_pts)
{
    float  pysize, pxsize;
    
    if (landscape_flag)  {  /* flip x and y */
      pysize = TOTAL_PXSIZE - (2.0*marg);
      pxsize = TOTAL_PYSIZE - (2.0*marg);  }
    else  {
      pxsize = TOTAL_PXSIZE - (2.0*marg);
      pysize = TOTAL_PYSIZE - (2.0*marg);  }
  /* see if x or y goes to edge; get scale factor, x-y center-margins */
    if (ysize/xsize > pysize/pxsize)  {
      *units_to_pts = pysize/ysize;   /* units times this => points */
      *x_center_marg = (pxsize - xsize*(*units_to_pts))/2.0;
      *y_center_marg = 0.0;  }
    else  {
      *units_to_pts = pxsize/xsize;
      *x_center_marg = 0.0;
      *y_center_marg = (pysize - ysize*(*units_to_pts))/2.0;  }
}

void ps_header(FILE *fpout, int landscape_flag, int pages)
{
  fprintf(fpout,"%%!PS-Adobe-2.0\n");
  if (pages>1)  {
    fprintf(fpout,"%%%%Pages: (atend)\n");
    fprintf(fpout,"%%%%EndComments\n");
    fprintf(fpout,"%%%%Page: 1 1\n");
  }
  fprintf(fpout,"%%%%BoundingBox: 0.0 0.0 612.0 792.0\n");
  /***fprintf(fpout,"initgraphics\n");***/
  fprintf(fpout,"1.00 1.00 scale\n");
  fprintf(fpout,"newpath\n");
  fprintf(fpout,"0.00 0.00 translate\n");
  if (landscape_flag) {
    fprintf(fpout,"90.00 rotate\n");
    fprintf(fpout,"0.00 -612.00 translate\n");
  }

  /* PostScript defs for a smaller output file */
  fprintf(fpout,"/c { closepath } def\n");
  fprintf(fpout,"/f { fill } def\n");
  fprintf(fpout,"/g { setgray } def\n");
  fprintf(fpout,"/l { lineto } def\n");
  fprintf(fpout,"/m { moveto } def\n");
  fprintf(fpout,"/n { newpath } def\n");
  fprintf(fpout,"/r { rotate } def\n");
  fprintf(fpout,"/s { stroke } def\n");
  fprintf(fpout,"/S { show } def\n");
  fprintf(fpout,"/w { setlinewidth } def\n");
  fprintf(fpout,"/fh { /Helvetica findfont } def\n");
  fprintf(fpout,"/ss { scalefont setfont } def\n");
}

void ps_parms(FILE *fpout, int argc, char **argv)
{
  int   i;
  char  line[200];
  FILE  *fpparm=NULL;

  fprintf(fpout,"%% cmdline:");
  for (i=0; i<argc; i++)
    fprintf(fpout," %s",argv[i]);
  fprintf(fpout,"\n");

  fpparm = fopen("arr.parm","r");
  if (fpparm==NULL) {
    fprintf(fpout,"%% no arr.parm: used all-default parms\n");
  }
  else {
    fprintf(fpout,"%% arr.parm (non-default parms):\n");
    while (fgets(line,80,fpparm) != NULL) {
      if (strlen(line) && line[0]!='\n') {
        fprintf(fpout,"%% %s",line);     /* fgets keeps newline */
        if (line[strlen(line)-1]!='\n')  /* if last missing */
          fprintf(fpout,"\n");
      }
    }
  }
}

void ps_nextpage(FILE *fpout, int nextpage)  /* call ready to write nextpage */
{
  fprintf(fpout,"showpage\n");
  fprintf(fpout,"%%%%Page: %d %d\n",nextpage,nextpage);
  fprintf(fpout,"%%%%BoundingBox: 0.0 0.0 612.0 792.0\n");
}

void ps_trailer(FILE *fpout, int pages)
{
  fprintf(fpout,"showpage\n");
  if (pages>1)  {
    fprintf(fpout,"%%%%Trailer\n");
    fprintf(fpout,"%%%%Pages: %d 1\n",pages);
  }
}

void ps_text(FILE *fpout, char *str, float x, float y, float fontsize)
{
  fprintf(fpout,"0.0 g\n");
  fprintf(fpout,"fh\n");
  fprintf(fpout,"%d ss\n",(int)fontsize);
  fprintf(fpout,"%3.2f %3.2f m\n",x,y);
  fprintf(fpout,"(%s) S\n",str);
}

void ps_pnt(FILE *fpout, float center_x, float center_y, float dia,
            float thickness, int filled_flag)
{
    fprintf(fpout,"n\n");
    fprintf(fpout,"0.0 g\n");
    fprintf(fpout,"%1.2f w\n",thickness);
    fprintf(fpout,"%3.2f %3.2f %3.2f 0.0 360.0 arc\n",
                    center_x,center_y,dia/2.0);
    if (filled_flag)
      fprintf(fpout,"f\n");
    fprintf(fpout,"s\n");
}

void ps_line(FILE *fpout, float x1, float y1, float x2, float y2,
             float thickness)
{
  fprintf(fpout,"n\n");
  fprintf(fpout,"%1.2f w\n",thickness);
  fprintf(fpout,"%3.2f %3.2f m\n",x1,y1);
  fprintf(fpout,"%3.2f %3.2f l\n",x2,y2);
  fprintf(fpout,"s\n");
}

void ps_box(FILE *fpout, float x0, float y0, float x1, float y1,
            float thickness)
{
  fprintf(fpout,"n\n");                    /* PS: newpath */
  fprintf(fpout,"0.0 g\n");                /* PS: setgray */
  fprintf(fpout,"%1.2f w\n",thickness);    /* PS: setlinewidth */
  fprintf(fpout,"%3.2f %3.2f m\n",x0,y0);  /* PS: moveto */
  fprintf(fpout,"%3.2f %3.2f l\n",x1,y0);  /* PS: lineto */
  fprintf(fpout,"%3.2f %3.2f l\n",x1,y1);  /* PS: lineto */
  fprintf(fpout,"%3.2f %3.2f l\n",x0,y1);  /* PS: lineto */
  fprintf(fpout,"c\n");                    /* PS: closepath */
  fprintf(fpout,"s\n");                    /* PS: stroke */
}

void ps_drawmesh(FILE *fpps, mesh_type mesh)
{
  int    x, y;
  float  marg, x_center_marg, y_center_marg, xmin, ymin, mm_to_pts, thickness;

  marg = mesh.marg;
  x_center_marg = mesh.x_center_marg;
  y_center_marg = mesh.y_center_marg;
  xmin = mesh.xmin;
  ymin = mesh.ymin;
  mm_to_pts = mesh.mm_to_pts;
  thickness = mesh.thickness;

  /* draw right and up vertices for each point */
  for (y=0; y<mesh.y_gridsize; y++)
  for (x=0; x<mesh.x_gridsize; x++) {
    if (x+1<mesh.x_gridsize)
      ps_line(fpps,
              marg + x_center_marg + (mesh.xm[y][x]-xmin)*mm_to_pts,
              marg + y_center_marg + (mesh.ym[y][x]-ymin)*mm_to_pts,
              marg + x_center_marg + (mesh.xm[y][x+1]-xmin)*mm_to_pts,
              marg + y_center_marg + (mesh.ym[y][x+1]-ymin)*mm_to_pts,
              thickness);
    if (y+1 < mesh.y_gridsize)
      ps_line(fpps,
              marg + x_center_marg + (mesh.xm[y][x]-xmin)*mm_to_pts,
              marg + y_center_marg + (mesh.ym[y][x]-ymin)*mm_to_pts,
              marg + x_center_marg + (mesh.xm[y+1][x]-xmin)*mm_to_pts,
              marg + y_center_marg + (mesh.ym[y+1][x]-ymin)*mm_to_pts,
              thickness);
  }  /* all mesh vertices */
}

void ps_draworigmeshbox(FILE *fpps, float gx0, float gy0, float gx1, float gy1,
                        mesh_type mesh)
{
  float  x0, y0, x1, y1;
  float  marg, x_center_marg, y_center_marg, xmin, ymin, mm_to_pts;

  marg = mesh.marg;
  x_center_marg = mesh.x_center_marg;
  y_center_marg = mesh.y_center_marg;
  xmin = mesh.xmin;
  ymin = mesh.ymin;
  mm_to_pts = mesh.mm_to_pts;
  
  x0 = marg + x_center_marg + (gx0-xmin)*mm_to_pts - 0.5*ORIGBBOX;
  y0 = marg + y_center_marg + (gy0-ymin)*mm_to_pts - 0.5*ORIGBBOX;
  x1 = marg + x_center_marg + (gx1-xmin)*mm_to_pts + 0.5*ORIGBBOX;
  y1 = marg + y_center_marg + (gy1-ymin)*mm_to_pts + 0.5*ORIGBBOX;
  ps_box(fpps, x0,y0, x1,y1, ORIGBBOX);
}

void ps_drawlesions(FILE *fpps, lesion_type *lesion, int num_lesion,
                    mesh_type mesh)
{
  int    i, x, y;
  float  marg, x_center_marg, y_center_marg, xmin, ymin, mm_to_pts,thickness;
  float  lesion_point;

  marg = mesh.marg;
  x_center_marg = mesh.x_center_marg;
  y_center_marg = mesh.y_center_marg;
  xmin = mesh.xmin;
  ymin = mesh.ymin;
  mm_to_pts = mesh.mm_to_pts;
  thickness = mesh.thickness;
  lesion_point = mesh.lesion_point;

  for (i=0; i<num_lesion; i++)  {
    ps_line(fpps,   /* line from start to target */
            marg + x_center_marg + (lesion[i].xstart-xmin)*mm_to_pts,
            marg + y_center_marg + (lesion[i].ystart-ymin)*mm_to_pts,
            marg + x_center_marg + (lesion[i].xtarg-xmin)*mm_to_pts,
            marg + y_center_marg + (lesion[i].ytarg-ymin)*mm_to_pts,
            5.0*thickness);
    ps_pnt(fpps,  /* start */
            marg + x_center_marg + (lesion[i].xstart-xmin)*mm_to_pts,
            marg + y_center_marg + (lesion[i].ystart-ymin)*mm_to_pts,
            lesion_point,0.2,0);
    ps_pnt(fpps,  /* target */
            marg + x_center_marg + (lesion[i].xtarg-xmin)*mm_to_pts,
            marg + y_center_marg + (lesion[i].ytarg-ymin)*mm_to_pts,
            lesion_point,0.2,1);
    x = lesion[i].xstart_vertex;
    y = lesion[i].ystart_vertex;
    ps_line(fpps,  /* line from start to target (nearest mesh point) */
            marg + x_center_marg + (mesh.xm[y][x]-xmin)*mm_to_pts,
            marg + y_center_marg + (mesh.ym[y][x]-ymin)*mm_to_pts,
            marg + x_center_marg + (lesion[i].xtarg_mesh-xmin)*mm_to_pts,
            marg + y_center_marg + (lesion[i].ytarg_mesh-ymin)*mm_to_pts,
            5.0*thickness);
    ps_pnt(fpps,  /* start mesh vertex (one being pulled on) */
            marg + x_center_marg + (mesh.xm[y][x]-xmin)*mm_to_pts,
            marg + y_center_marg + (mesh.ym[y][x]-ymin)*mm_to_pts,
            1.5*lesion_point,0.2,0);
    ps_pnt(fpps,  /* location of mesh vertex that puts start lesion on targ*/
            marg + x_center_marg + (lesion[i].xtarg_mesh-xmin)*mm_to_pts,
            marg + y_center_marg + (lesion[i].ytarg_mesh-ymin)*mm_to_pts,
            2.0*lesion_point,0.2,0);
  }
}

void ps_drawdata(FILE *fpps, rf_type *rf, int num_rf, mesh_type mesh,
                 int showstart_flag, int showfinal_flag)   /* points */
{
  int    i;
  float  marg, x_center_marg, y_center_marg, xmin, ymin, mm_to_pts, thickness;
  float  data_point;

  marg = mesh.marg;
  x_center_marg = mesh.x_center_marg;
  y_center_marg = mesh.y_center_marg;
  xmin = mesh.xmin;
  ymin = mesh.ymin;
  mm_to_pts = mesh.mm_to_pts;
  thickness = mesh.thickness;
  data_point = mesh.data_point;

  /* illustrate final (and original if asked) location of data points */
  for (i=0; i<num_rf; i++)  {
    if (showstart_flag && showfinal_flag)  {
      ps_line(fpps,  /* lines from start to final */
              marg + x_center_marg + (rf[i].xcortex-xmin)*mm_to_pts,
              marg + y_center_marg + (rf[i].ycortex-ymin)*mm_to_pts,
              marg + x_center_marg + (rf[i].final_xcortex-xmin)*mm_to_pts,
              marg + y_center_marg + (rf[i].final_ycortex-ymin)*mm_to_pts,
              thickness);
      ps_pnt(fpps,
              marg + x_center_marg + (rf[i].xcortex-xmin)*mm_to_pts,
              marg + y_center_marg + (rf[i].ycortex-ymin)*mm_to_pts,
              data_point,0.2,0);
    }
    if (showfinal_flag)
      ps_pnt(fpps,
              marg + x_center_marg + (rf[i].final_xcortex-xmin)*mm_to_pts,
              marg + y_center_marg + (rf[i].final_ycortex-ymin)*mm_to_pts,
              data_point,0.2,1);
  }  /* all data points */
}

void ps_drawlegend(FILE *fpps, char *filename, int landscape_flag, int steps,
                   float sqfactor, float crosspnt, float set_data_dist_term,
                   float set_init_dist_term, float set_ave_dist_term,
                   float set_orth_term, float set_lin_term, float momentum,
                   float epsilon, int maxsteps, int frac_steps_to_anneal,
                   float frac_epsilon_to_anneal, mesh_type mesh)
{
  int    x_gridsize, y_gridsize;
  float  ydraw, line;
  char   str[200];
  float  marg, mm_to_pts;

  marg = mesh.marg;
  mm_to_pts = mesh.mm_to_pts;
  x_gridsize = mesh.x_gridsize;
  y_gridsize = mesh.y_gridsize;

  if (landscape_flag) ydraw = TOTAL_PXSIZE-marg;
  else                ydraw = TOTAL_PYSIZE-marg;
  line = 0.0;
  sprintf(str,"file:  %s",filename);
  ps_text(fpps,str,marg,ydraw-line*LEGEND_FONT,LEGEND_FONT);
  line += 1.0;
  sprintf(str,"  %d relaxation steps\n",steps);
  ps_text(fpps,str,marg,ydraw-line*LEGEND_FONT,LEGEND_FONT);
  if ((int)((float)maxsteps*frac_steps_to_anneal) > 0)  {
    line += 1.0;
    sprintf(str,"  %d/%d steps anneal at %g epsilon\n",
                      (int)((float)maxsteps*frac_steps_to_anneal),
                      maxsteps,frac_epsilon_to_anneal);
    ps_text(fpps,str,marg,ydraw-line*LEGEND_FONT,LEGEND_FONT);
  }
  line += 1.0;
  sprintf(str,"  %d x %d mesh points\n",x_gridsize,y_gridsize);
  ps_text(fpps,str,marg,ydraw-line*LEGEND_FONT,LEGEND_FONT);
  line += 1.0;
  if (sqfactor>0.0) {
    sprintf(str,"  sqfactor = %g\n",sqfactor);
    ps_text(fpps,str,marg,ydraw-line*LEGEND_FONT,LEGEND_FONT);
    line += 1.0;
    sprintf(str,"  crosspnt = %g\n",crosspnt);
    ps_text(fpps,str,marg,ydraw-line*LEGEND_FONT,LEGEND_FONT);
  }
  else {
    sprintf(str,"  use actual dist,angles\n");
    ps_text(fpps,str,marg,ydraw-line*LEGEND_FONT,LEGEND_FONT);
  }
  line += 1.0;
  sprintf(str,"  data term = %2.1f\n",set_data_dist_term);
  ps_text(fpps,str,marg,ydraw-line*LEGEND_FONT,LEGEND_FONT);
  line += 1.0;
  sprintf(str,"  init term = %g\n",set_init_dist_term);
  ps_text(fpps,str,marg,ydraw-line*LEGEND_FONT,LEGEND_FONT);
  line += 1.0;
  sprintf(str,"  ave term = %g\n",set_ave_dist_term);
  ps_text(fpps,str,marg,ydraw-line*LEGEND_FONT,LEGEND_FONT);
  line += 1.0;
  sprintf(str,"  orth term = %g\n",set_orth_term);
  ps_text(fpps,str,marg,ydraw-line*LEGEND_FONT,LEGEND_FONT);
  line += 1.0;
  sprintf(str,"  lin term = %g\n",set_lin_term);
  ps_text(fpps,str,marg,ydraw-line*LEGEND_FONT,LEGEND_FONT);
  line += 1.0;
  sprintf(str,"  momentum = %g\n",momentum);
  ps_text(fpps,str,marg,ydraw-line*LEGEND_FONT,LEGEND_FONT);
  line += 1.0;
  sprintf(str,"  epsilon = %g mm\n",epsilon);
  ps_text(fpps,str,marg,ydraw-line*LEGEND_FONT,LEGEND_FONT);

  /* write cortex-coords scale bar w/centered (1.33) text below legend */
  line += 1.0;
  ps_text(fpps,
          "1 mm",
           marg + 2.0*LEGEND_FONT + (1.0*mm_to_pts)/2.0 - 1.33*LEGEND_FONT,
           ydraw - line*LEGEND_FONT,
           LEGEND_FONT);
    ps_line(fpps,
          marg + 2.0*LEGEND_FONT,             /* start a few chars in */
          ydraw - line*LEGEND_FONT - 3.0,     /* line close underneath */
          marg + 2.0*LEGEND_FONT + 1.0*mm_to_pts,
          ydraw - line*LEGEND_FONT - 3.0,
          2.0);
}

/*============================================================================
   Interpolate data points using four vertices surrounding each point.
============================================================================*/
void interpolate_data(rf_type *rf, int num_rf, mesh_type mesh)
{
  int     i, x, y;
  float   xcortex=0.0, ycortex=0.0;
  float   xoffset, yoffset;
  float   adj_xoffset, adj_yoffset;

  for (i=0; i<num_rf; i++) {
    xoffset = rf[i].xoffset;
    yoffset = rf[i].yoffset;
    adj_xoffset = -copysign((mesh.x_incr - fabs(xoffset)),xoffset);/*gees..*/
    adj_yoffset = -copysign((mesh.y_incr - fabs(yoffset)),yoffset);
    x = rf[i].xvertex;
    y = rf[i].yvertex;
    if (xoffset>=0.0 && yoffset>=0.0) {  /* first quadrant */
      xcortex = (mesh.xm[y][x] + xoffset +
                 mesh.xm[y][x+1] + adj_xoffset +
                 mesh.xm[y+1][x+1] + adj_xoffset +
                 mesh.xm[y+1][x] + xoffset)/4.0;  /* counterclockwise */
      ycortex = (mesh.ym[y][x] + yoffset +
                 mesh.ym[y][x+1] + yoffset +
                 mesh.ym[y+1][x+1] + adj_yoffset +
                 mesh.ym[y+1][x] + adj_yoffset)/4.0;
    }
    if (xoffset<0.0 && yoffset>=0.0) {  /* second quadrant */
      xcortex = (mesh.xm[y][x] + xoffset +
                 mesh.xm[y+1][x] + xoffset +
                 mesh.xm[y+1][x-1] + adj_xoffset +
                 mesh.xm[y][x-1] + adj_xoffset)/4.0;
      ycortex = (mesh.ym[y][x] + yoffset +
                 mesh.ym[y+1][x] + adj_yoffset +
                 mesh.ym[y+1][x-1] + adj_yoffset +
                 mesh.ym[y][x-1] + yoffset)/4.0;
    }
    if (xoffset<0.0 && yoffset<0.0) {  /* third quadrant */
      xcortex = (mesh.xm[y][x] + xoffset +
                 mesh.xm[y][x-1] + adj_xoffset +
                 mesh.xm[y-1][x-1] + adj_xoffset +
                 mesh.xm[y-1][x] + xoffset)/4.0;
      ycortex = (mesh.ym[y][x] + yoffset +
                 mesh.ym[y][x-1] + yoffset +
                 mesh.ym[y-1][x-1] + adj_yoffset +
                 mesh.ym[y-1][x] + adj_yoffset)/4.0;
    }
    if (xoffset>=0.0 && yoffset<0.0) {  /* fourth quadrant */
      xcortex = (mesh.xm[y][x] + xoffset +
                 mesh.xm[y-1][x] + xoffset +
                 mesh.xm[y-1][x+1] + adj_xoffset +
                 mesh.xm[y][x+1] + adj_xoffset)/4.0;
      ycortex = (mesh.ym[y][x] + yoffset +
                 mesh.ym[y-1][x] + adj_yoffset +
                 mesh.ym[y-1][x+1] + adj_yoffset +
                 mesh.ym[y][x+1] + yoffset)/4.0;
    }
    rf[i].final_xcortex = xcortex;   /* save in structure */
    rf[i].final_ycortex = ycortex;
  }  /* all data file points */
}

float **fmatrix(int ysize, int xsize)
{
  int    i;
  float  **array;

  array = (float **)malloc(ysize*sizeof(float*));
  if (!array) printf("stretch: ### allocation failure 1 in fmatrix()");
  for(i = 0; i < ysize; i++)  {
    array[i] = (float *)malloc(xsize*sizeof(float));
    if (!array[i]) printf("stretch: ### allocation failure 2 in fmatrix()");
  }
  return array;
}

void shuffle(int *array, int n)
{
  int  i,j,tmp;

  for (i=0;i<n-1;i++)  {
    j = random()%(n-i) + i;
    tmp = array[j];
    array[j] = array[i];
    array[i] = tmp;
  }
}

/*============================================================================
  Get energy using 4 surrounding vertices (N.B.: array of lesion_type's,
  but only one copy of mesh_type).
============================================================================*/
void get_energy_terms(lesion_type *lesion, int num_lesion, mesh_type *mesh,
                      int x, int y, float xeps, float yeps, int release_flag)
{
  int    i, icnt;
  int    cnt0, cnt1, cnt2, cnt3;
  float  dx, dy;
  float  dx0=0.0, dx1=0.0, dx2=0.0, dx3=0.0;
  float  dy0=0.0, dy1=0.0, dy2=0.0, dy3=0.0;
  float  dist0, dist1, dist2, dist3;
  float  th0, th1, th2, th3;
  float  dth0=0.0, dth1=0.0, dth2=0.0, dth3=0.0;
  float  ddth0, ddth1, ddth2, ddth3;
  float  ave_dist;
  float  sum;
  float  cnt;

  /*** DATA LENGTH term: if anchor point, get data term, else leave zero */
  mesh->data = 0.0;
  if (!release_flag) {  /* skip if released from mesh */
    for (i=0; i<num_lesion; i++)  {
      if (x==lesion[i].xstart_vertex && y==lesion[i].ystart_vertex)  {
        dx = lesion[i].xtarg_mesh - (mesh->xm[y][x]+xeps);
        dy = lesion[i].ytarg_mesh - (mesh->ym[y][x]+yeps);
        if (fabs(dx)>DEPS && fabs(dy)>DEPS)
          mesh->data = hypot(dx,dy);
      }
    }  /* for all lesions */
  }  /* not released from data */

  /*** INIT LENGTH term: difference from original length */
  cnt0 = cnt1 = cnt2 = cnt3 = 0;
  dist0 = dist1 = dist2 = dist3 = 0.0;
  /* get dist to E,N,W,S mesh pts (retain proper sign for dx,dy for atan2) */
  if (x+1 < mesh->x_gridsize)  {
    dx0 = mesh->xm[y][x+1] - (mesh->xm[y][x]+xeps);  /* east */
    dy0 = mesh->ym[y][x+1] - (mesh->ym[y][x]+yeps);
    if (fabs(dx0)>DEPS && fabs(dy0)>DEPS) {
      dist0 = hypot(dx0,dy0);
      cnt0++;
    }
  }
  if (y+1<mesh->y_gridsize) {
    dx1 = mesh->xm[y+1][x] - (mesh->xm[y][x]+xeps);  /* north */
    dy1 = mesh->ym[y+1][x] - (mesh->ym[y][x]+yeps);
    if (fabs(dx1)>DEPS && fabs(dy1)>DEPS) {
      dist1 = hypot(dx1,dy1);
      cnt1++;
    }
  }
  if (x-1>=0)  {
    dx2 = mesh->xm[y][x-1] - (mesh->xm[y][x]+xeps);  /* west */
    dy2 = mesh->ym[y][x-1] - (mesh->ym[y][x]+yeps);
    if (fabs(dx2)>DEPS && fabs(dy2)>DEPS) {
      dist2 = hypot(dx2,dy2);
      cnt2++;
    }
  }
  if (y-1 >= 0)  {
    dx3 = mesh->xm[y-1][x] - (mesh->xm[y][x]+xeps);  /* south */
    dy3 = mesh->ym[y-1][x] - (mesh->ym[y][x]+yeps);
    if (fabs(dx3)>DEPS && fabs(dy3)>DEPS) {
      dist3 = hypot(dx3,dy3);
      cnt3++;
    }
  }
  icnt = cnt0 + cnt1 + cnt2 + cnt3;
  if (icnt==0) {
    mesh->init = mesh->ave = mesh->orth = mesh->lin = 0.0;
    return;
  }
  cnt = (float)icnt;
  sum = 0.0;
  if (cnt0)  sum += fabs(dist0 - mesh->x_incr);
  if (cnt1)  sum += fabs(dist1 - mesh->x_incr);
  if (cnt2)  sum += fabs(dist2 - mesh->x_incr);
  if (cnt3)  sum += fabs(dist3 - mesh->x_incr);
  mesh->init = sum/cnt;
  if (mesh->sqfactor > 0.0)  { /* power term: set lin crossover,slope */
    mesh->init += mesh->sqfactor *
       ((CUBE(mesh->init))/mesh->x_incr/(SQR(mesh->crosspnt)) - mesh->init);
  }
  /*** AVE LENGTH term: difference from average length */
  sum = 0.0;
  ave_dist = (dist0 + dist1 + dist2 + dist3)/cnt;
  if (cnt0)  sum += fabs(dist0 - ave_dist);
  if (cnt1)  sum += fabs(dist1 - ave_dist);
  if (cnt2)  sum += fabs(dist2 - ave_dist);
  if (cnt3)  sum += fabs(dist3 - ave_dist);
  mesh->ave = sum/cnt;
  if (mesh->sqfactor > 0.0)  {  /* should use starting dist again? */
    mesh->ave += mesh->sqfactor *
             ((CUBE(mesh->ave))/ave_dist/(SQR(mesh->crosspnt)) - mesh->ave);
  }
  /*** AVE NEIGHBORS LENGTH term: (not implemented) */
  /*** CONFORMALITY term: difference from orthogonality */
  th0 = th1 = th2 = th3 = 0.0;
  ddth0 = ddth1 = ddth2 = ddth3 = 0.0;
  if (cnt0)  th0 = atan2(dy0,dx0);
  if (cnt1)  th1 = atan2(dy1,dx1);
  if (cnt2)  th2 = atan2(dy2,dx2);
  if (cnt3)  th3 = atan2(dy3,dx3);
  /* get angle for each pair of vertices (catch wrap) */
  if (cnt0 && cnt1)  {
    if (th1 - th0 >= M_PI)  th0 += 2.0*M_PI;
    if (th1 - th0 < -M_PI)  th1 += 2.0*M_PI;
    dth0 = fabs(th1 - th0);       /* angle between two vertices */
    ddth0 = fabs(dth0 - M_PI_2);  /* difference from orthogonal */
  }
  if (cnt1 && cnt2)  {
    if (th2 - th1 >= M_PI)  th1 += 2.0*M_PI;
    if (th2 - th1 < -M_PI)  th2 += 2.0*M_PI;
    dth1 = fabs(th2 - th1);
    ddth1 = fabs(dth1 - M_PI_2);
  }
  if (cnt2 && cnt3)  {
    if (th3 - th2 >= M_PI)  th2 += 2.0*M_PI;
    if (th3 - th2 < -M_PI)  th3 += 2.0*M_PI;
    dth2 = fabs(th3 - th2);
    ddth2 = fabs(dth2 - M_PI_2);
  }
  if (cnt3 && cnt0)  {
    if (th0 - th3 >= M_PI)  th3 += 2.0*M_PI;
    if (th0 - th3 < -M_PI)  th0 += 2.0*M_PI;
    dth3 = fabs(th0 - th3);
    ddth3 = fabs(dth3 - M_PI_2);
  }
  mesh->orth = (ddth0 + ddth1 + ddth2 + ddth3)/cnt;
  if (mesh->sqfactor > 0.0)  {
    mesh->orth += mesh->sqfactor *
             ((CUBE(mesh->orth))/M_PI_2/(SQR(mesh->crosspnt)) - mesh->orth);
  }
  /*** LINEARITY term: difference from linear across vertex */
  ddth0 = ddth1 = cnt = 0.0;
  if (cnt0 && cnt2)  {  /* horizontal */
    if (!cnt1)       ddth0 = fabs(dth2+dth3 - M_PI);  /* upper edge */
    else if (!cnt3)  ddth0 = fabs(dth0+dth1 - M_PI);  /* lower edge */
    else             ddth0 = fabs(dth0+dth1 - M_PI);  /* inside */
    cnt += 1.0;
  }
  if (cnt1 && cnt3)  {  /* vertical */
    if (!cnt0)       ddth1 = fabs(dth1+dth2 - M_PI);  /* right edge */
    else if (!cnt2)  ddth1 = fabs(dth3+dth0 - M_PI);  /* left edge */
    else             ddth1 = fabs(dth3+dth0 - M_PI);  /* inside */
    cnt += 1.0;
  }
  if (cnt==0.0)  cnt = 1.0;    /* fix div by zero on corner pnts, no lin */
  mesh->lin = (ddth0 + ddth1)/cnt;
  if (mesh->sqfactor>0.0)  {
    mesh->lin += mesh->sqfactor *
              ((CUBE(mesh->lin))/M_PI_2/(SQR(mesh->crosspnt)) - mesh->lin);
  }
}

/*============================================================================
  Hack replacement for drem missing on HP
============================================================================*/
double mydrem(double x, double y)
{
  double  a, n, fl, ce;
  
  a = x/y;
  fl = floor(a);
  ce = ceil(a);
  if (ce-a > a-fl) n = fl;
  else             n = ce;
  return(x - n*y);
}
/*============================================================================
  Hack replacement for copysign missing on HP
============================================================================*/
double mycopysign(double x, double y)
{
  if (y>=0.0)  return(fabs(x));
  else         return(-fabs(x));
}

/*============================================================================
  Print help
============================================================================*/
void print_help(void)
{
printf("\
#############################################################################\n\
              stretch -- deform sparse x-y array of points\n\
\n\
                       written by Marty Sereno\n\
                  University of California, San Diego\n\
#############################################################################\n\
    Stretches sparsely sampled x-y array of cortical recording points\n\
  using a set of marker lesions for points identified after physically \n\
  flattening the cortex.  Designed to be used before data is interpolated\n\
  with interpolate.c or re-drawn with arrows.c.  Data points or arrows or\n\
  interpolated contour plot can then be superimposed on myelin photograph.\n\
\n\
    Reads two ASCII data files (rootfilename.out, rootfilename-les.out)\n\
  and writes out ASCII datafile (rootfilename.str) suitable for arrows.c\n\
  (again) or interpolate.c, and one or more PostScript representations of\n\
  selected steps in the stretching process:\n\
  \n\
Infiles:\n\
  rootfilename.out         8 args/line:  name x y r theta len width angle\n\
  rootfilename-les.out     [same format]\n\
\n\
Outfiles:\n\
  rootfilename.str         [same format]\n\
  rootfilename.str.#.ps    PostScript files for requested steps\n\
  \n\
    Input file is RF params for data sparsely distributed in the x-y plane.\n\
  Same format as for arrows.c and rfplots.c, but with no comments or 3-arg\n\
  parms.  All transformations (penetration coords rotation; units -> mm\n\
  scaling; RF coord transformation; flip hemifield, flip hemisphere) should\n\
  be done by arrows.c.  *Both* the input file and the lesions file must be\n\
  transformed with same parms.  Output file is same as input file except that\n\
  x-y coords have been stretched to line up the lesion points with their\n\
  actual x-y location measured from the flatmount.  The names of the points\n\
  in the lesion file must be the same as their counterparts in the input\n\
  datafile.\n\
\n\
    PostScript files can be generated for intermediate stages in the\n\
  stretching process to help in optimizing parameters.  Control lesion\n\
  points are always illustrated (start, target, meshstart, meshtarget).\n\
  Optionally, the mesh, the final data point locations, and the starting\n\
  point locations (and a line from start to final) can be illustrated\n\
  by setting the SHOWMESH, SHOWSTART, and SHOWFINAL flags in the parmfile.\n\
\n\
    Also reads parm file (str.parm) file containing various settings\n\
  (e.g., coefficients on energy function terms, x_gridsize); if not there,\n\
  writes a default version of the file, which can then be edited.  Ignores\n\
  properly formatted parameters in parm file that it doesn't recognize.\n\
  If a parm appears more than once, the last instance determines its value.\n\
  Parms omitted from file are set to default values.  Default parm file:\n\
\n\
MARG = 30.0\n\
THICKNESS = 0.2\n\
DATA_POINT = 3.0\n\
LESION_POINT = 7.0\n\
LANDSCAPE_FLAG = 0\n\
LEGEND_FLAG = 1\n\
SHOWMESH_FLAG = 1\n\
SHOWSTART_FLAG = 1\n\
SHOWFINAL_FLAG = 1\n\
DATA_MARG_FRAC = 0.02\n\
X_GRIDSIZE = 20\n\
SQFACTOR = 1.0\n\
CROSSPNT = 0.25\n\
SET_DATA_DIST_TERM = 1.0\n\
SET_INIT_DIST_TERM = 0.01\n\
SET_AVE_DIST_TERM = 0.1\n\
SET_ORTH_TERM = 0.1\n\
SET_LIN_TERM = 0.0\n\
MOMENTUM = 0.5\n\
EPSILON = 0.1\n\
FRAC_STEPS_TO_ANNEAL = 0.0\n\
FRAC_EPSILON_TO_ANNEAL = 0.25\n\
FRAC_STEPS_TO_RELEASE = 0.0\n\
\n\
    The five energy function terms are calculated using the distances to\n\
  the four surrounding vertices and the angles between the edges to those\n\
  neigbors (positive is bad):\n\
\n\
  DATA_DIST: dist between lesion in unstretched file and lesion file\n\
  INIT_DIST: ave diff from starting distance to N,S,E,W vertices\n\
  AVE_DIST:  ave diff from average distance to N,S,E,W vertices\n\
  ORTH:      ave diff from orthogonality (all four angles)\n\
  LIN:       ave diff from N-S and E-W linearity across a vertex\n\
\n\
    Setting SQFACTOR adjusts how much the term approaches the term^3.  It\n\
  is scaled so that the point at which the power term goes past f(x)=x is\n\
  set by CROSSPNT, which is the fraction of MAX (which is either the starting\n\
  mesh element size or average size for distance terms, and fraction of 90\n\
  degrees for angle terms):\n\
      term = term + SQFACTOR*(term^3/MAX/CROSSPNT^2 - term).\n\
  SQFACTOR = 1.0 makes term go up like x^3; less than one makes it closer\n\
  to linear.  This adjusts the shape of the increased penalty for big errors\n\
  and reduced penalty for small errors.  If SQFACTOR>1.0, it is automatically\n\
  reset to 1.0.  Higher powers behave quite similarly.\n\
\n\
    Set LANDSCAPE_FLAG to get landscape output (this is only transformation\n\
  not done in *.out files).\n\
##########################################################################\n\
\n\
");
}

